/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_ATTACHEDFUNCTION_FUNCTION_HH
#define DUNE_FEM_ATTACHEDFUNCTION_FUNCTION_HH

#include <dune/fem/space/common/dofmanager.hh>
#include <dune/fem/function/common/dofblock.hh>
#include <dune/fem/function/common/discretefunction.hh>
#include <dune/fem/function/localfunction/standardlocalfunction.hh>

#include "container.hh"

namespace Dune
{

  namespace Fem
  {

  template< class DiscreteFunctionSpace >
  class AttachedDiscreteFunction;

  
  template< class DiscreteFunctionSpace >
  struct AttachedDiscreteFunctionTraits
  {
    typedef DiscreteFunctionSpace DiscreteFunctionSpaceType;

    typedef AttachedDiscreteFunction< DiscreteFunctionSpaceType >
      DiscreteFunctionType;
    typedef AttachedDiscreteFunctionTraits< DiscreteFunctionSpaceType >
      Traits;

    typedef StandardLocalFunctionFactory< Traits > LocalFunctionFactoryType;
    typedef LocalFunctionStack< LocalFunctionFactoryType >
      LocalFunctionStorageType;
    typedef typename LocalFunctionStorageType :: LocalFunctionType
      LocalFunctionType;
    
    typedef typename DiscreteFunctionSpaceType :: DomainFieldType
      DomainFieldType;
    typedef typename DiscreteFunctionSpaceType :: RangeFieldType RangeFieldType;
    typedef typename DiscreteFunctionSpaceType :: DomainType DomainType;
    typedef typename DiscreteFunctionSpaceType :: RangeType RangeType;
    typedef typename DiscreteFunctionSpaceType :: JacobianRangeType
      JacobianRangeType;

    typedef typename DiscreteFunctionSpaceType :: MapperType MapperType;
    typedef typename DiscreteFunctionSpaceType :: GridPartType GridPartType;

    typedef typename GridPartType :: GridType GridType;

    typedef RangeFieldType DofType;
    typedef AttachedDiscreteFunctionContainer< DofType, GridType, MapperType >
      ContainerType;

    typedef typename ContainerType :: SlotIteratorType DofIteratorType;
    typedef typename ContainerType :: ConstSlotIteratorType
      ConstDofIteratorType;

    enum { blockSize = DiscreteFunctionSpaceType :: localBlockSize };

    typedef DofBlockProxy< DiscreteFunctionType, DofType, blockSize >
      DofBlockType;
    typedef DofBlockProxy
      < const DiscreteFunctionType, const DofType, blockSize >
      ConstDofBlockType;
    typedef Fem::Envelope< DofBlockType > DofBlockPtrType;
    typedef Fem::Envelope< ConstDofBlockType > ConstDofBlockPtrType;
  };



  template< class DiscreteFunctionSpace >
  class AttachedDiscreteFunction
  : public DiscreteFunctionDefault
      < AttachedDiscreteFunctionTraits< DiscreteFunctionSpace > >
  {
    typedef AttachedDiscreteFunction< DiscreteFunctionSpace > ThisType;
    typedef DiscreteFunctionDefault
      < AttachedDiscreteFunctionTraits< DiscreteFunctionSpace > >
      BaseType;

  public:
    typedef AttachedDiscreteFunctionTraits< DiscreteFunctionSpace > Traits;

    typedef typename Traits :: DiscreteFunctionSpaceType
      DiscreteFunctionSpaceType;

    typedef typename Traits :: DomainFieldType DomainFieldType;
    typedef typename Traits :: RangeFieldType RangeFieldType;

    typedef typename Traits :: DomainType DomainType;
    typedef typename Traits :: RangeType RangeType;

    typedef typename Traits :: LocalFunctionFactoryType
      LocalFunctionFactoryType;

    typedef typename Traits :: ContainerType ContainerType;

    typedef typename Traits :: DofIteratorType DofIteratorType;
    typedef typename Traits :: ConstDofIteratorType ConstDofIteratorType;

    typedef typename Traits :: DofType DofType;
    typedef typename Traits :: DofBlockType DofBlockType;
    typedef typename Traits :: ConstDofBlockType ConstDofBlockType;
    typedef typename Traits :: DofBlockPtrType DofBlockPtrType;
    typedef typename Traits :: ConstDofBlockPtrType ConstDofBlockPtrType;

  protected:
    const LocalFunctionFactoryType lfFactory_;

    ContainerType &container_;
    unsigned int slot_;

  public:
    inline AttachedDiscreteFunction ( const std :: string &name,
                                      const DiscreteFunctionSpaceType &dfSpace )
    : BaseType( name, dfSpace, lfFactory_ ),
      lfFactory_( *this ),
      container_( ContainerType :: attach( dfSpace.grid(), dfSpace.mapper() ) ),
      slot_( container_.allocSlot() )
    {}

    inline AttachedDiscreteFunction ( const ThisType &other )
    : BaseType( other.name(), other.space(), lfFactory_ ),
      lfFactory_( *this ),
      container_( ContainerType :: attach
        ( other.space().grid(), other.space().mapper() ) ),
      slot_( container_.allocSlot() )
    {
      assign( other );
    }

    inline ~AttachedDiscreteFunction ()
    {
      container_.freeSlot( slot_ );
      ContainerType :: detach( container_ );
    }

    inline ConstDofIteratorType dbegin () const
    {
      return container().begin( slot_ );
    }

    inline DofIteratorType dbegin ()
    {
      return container().begin( slot_ );
    }

    inline ConstDofIteratorType dend () const
    {
      return container().end( slot_ );
    }

    inline DofIteratorType dend ()
    {
      return container().end( slot_ );
    }

    inline ConstDofBlockPtrType block ( unsigned int index ) const
    {
      typename ConstDofBlockType :: KeyType key( this, index );
      return ConstDofBlockPtrType( key );
    }

    inline DofBlockPtrType block ( unsigned int index )
    {
      typename DofBlockType :: KeyType key( this, index );
      return DofBlockPtrType( key );
    }

    inline const RangeFieldType &dof ( unsigned int index ) const
    {
      return container().dof( slot_, index );
    }

    inline RangeFieldType &dof ( unsigned int index )
    {
      return container().dof( slot_, index );
    }

    inline void enableDofCompression ()
    {
      container().enableDofCompression();
    }

    inline int size () const
    {
      return container_.size();
    }

    inline void swap ( ThisType &other )
    {
      const unsigned int myslot = slot_;
      slot_ = other.slot_;
      other.slot_ = myslot;
    }

  protected:
    inline const ContainerType &container () const
    {
      return container_;
    }

    inline ContainerType &container ()
    {
      return container_;
    }
  };

  } // namespace Fem

  // #if DUNE_FEM_COMPATIBILITY  
  // put this in next version 1.4 

  using Fem :: AttachedDiscreteFunction ;
  // #endif // DUNE_FEM_COMPATIBILITY

} // namespace Dune

#endif
