/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GENERICLOCALFUNCTION_INLINE_HH
#define DUNE_FEM_GENERICLOCALFUNCTION_INLINE_HH

#include "genericlocalfunction.hh"

namespace Dune
{

  namespace Fem
  {

  // GenericLocalFunctionImpl
  // -------------------------

  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::GenericLocalFunctionImpl ( DiscreteFunctionType &discreteFunction )
  : discreteFunction_( discreteFunction ),
    values_( DiscreteFunctionSpace::localBlockSize * discreteFunction_.space().blockMapper().maxNumDofs() ),
    baseFunctionSet_(),
    entity_( 0 ),
    numDofs_( 0 ),
    needCheckGeometry_( true )
  {}
 


  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::GenericLocalFunctionImpl ( const ThisType &other )
  : discreteFunction_( other.discreteFunction_ ),
    values_( other.values_ ),
    baseFunctionSet_( other.baseFunctionSet_ ),
    entity_( other.entity_ ),
    numDofs_( other.numDofs_ ),
    needCheckGeometry_( other.needCheckGeometry_ )
  {}



  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline
  const typename GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::RangeFieldType &
  GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::operator[] ( const int num ) const
  {
    assert( (num >= 0) && (num < numDofs()) );
    return *(values_[ num ]);
  }
  

  
  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline
  typename GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::RangeFieldType &
  GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::operator[] ( const int num )
  {
    assert( (num >= 0) && (num < numDofs()) );
    return *(values_[ num ]);
  }



  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline int
  GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::order () const
  {
    return discreteFunction_.space().order( entity() );
  }



  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline
  const typename GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::BaseFunctionSetType &
  GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::baseFunctionSet () const
  {
    assert( entity_ != 0 );
    return baseFunctionSet_;
  }


  
  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline
  const typename GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::EntityType &
  GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::entity () const
  {
    assert( entity_ != 0 );
    return *entity_;
  }



  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline void GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >::init ( const EntityType &entity )
  {
    typedef typename DiscreteFunctionSpaceType::MapperType MapperType;
   
    const DiscreteFunctionSpaceType &space = discreteFunction_.space();
    const bool multipleBaseSets = space.multipleBaseFunctionSets();

    if( multipleBaseSets || needCheckGeometry_ )
    {
      // if multiple base sets skip geometry call
      bool updateBaseSet = true;
      if( !multipleBaseSets && (entity_ != 0) )
        updateBaseSet = (baseFunctionSet_.geometryType() != entity.type());
      
      if( multipleBaseSets || updateBaseSet )
      {
        baseFunctionSet_ = space.baseFunctionSet( entity );

        // note, do not use baseFunctionSet() here, entity might no have been set
        numDofs_ = baseFunctionSet_.size();

        needCheckGeometry_ = space.multipleGeometryTypes();
      }
    }

    // cache entity
    entity_ = &entity;
    assert( baseFunctionSet_.geometryType() == entity.type() );

    assert( numDofs_ <= values_.size() );
    const MapperType &mapper = space.mapper();

    mapper.map( entity, indices_ );
    assert( indices_.size() == numDofs_ );
    for( unsigned int i = 0; i < numDofs_; ++i )
      values_[ i ] = &discreteFunction_.dof( indices_[ i ] );
  }



  template< class DiscreteFunction, class DiscreteFunctionSpace >
  inline int GenericLocalFunctionImpl< DiscreteFunction, DiscreteFunctionSpace >
    ::numDofs () const
  {
    return numDofs_;
  }

  } // namespace Fem

} // namespace Dune

#endif // DUNE_FEM_GENERICLOCALFUNCTION_INLINE_HH
