/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_PRODUCTFUNCTION_HH
#define DUNE_PRODUCTFUNCTION_HH

//- system includes 
#include <fstream>

//- Dune includes 
#if HAVE_DUNE_GEOMETRY
#include <dune/geometry/type.hh>
#else
#include <dune/common/geometrytype.hh>
#endif

#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/function/common/dofiterator.hh>
#include <dune/fem/space/common/dofmanager.hh>

namespace Dune{

namespace Fem 
{

  template <class DiscreteFunctionSpaceImp,class DiscreteFunctionSpace2Imp> class ProductDiscreteFunction;

  template <class DiscreteFunctionSpaceImp>
  struct ProductDiscreteFunctionTraits {
    typedef DiscreteFunctionSpaceImp DiscreteFunctionSpaceType;
  
    typedef ProductDiscreteFunction<DiscreteFunctionSpaceImp,DiscreteFunctionSpaceImp> DiscreteFunctionType;
    typedef MutableArray<typename DiscreteFunctionSpaceImp::RangeFieldType> DofArrayType;
      
    typedef typename DofArrayType::DofIteratorType DofIteratorType;
    typedef typename DofArrayType::ConstDofIteratorType ConstDofIteratorType;
  };

//**********************************************************************
//! @addtogroup ProductDFunction
// --ProductDiscreteFunction 
//
//! \brief This is one special implementation of a discrete function  \f$  u(x,y) = \sum _{i=1} ^{n_x}\sum _{j=1} ^{n_y} u_{ij}\varphi _i(x) \psi _j(y) \f$ for product spaces \f$ X \times Y \f$ using an
//! array for storing the dofs with \f$u_{ij}= u\left[j n_x +i\right]\f$.  
//!
//!
//**********************************************************************
template<class DiscreteFunctionSpaceType, class DiscreteFunctionSpace2Type>
class ProductDiscreteFunction 
{
public:
  //!Type of range field, (usually a float type)
  typedef typename DiscreteFunctionSpaceType::RangeFieldType RangeFieldType;
  typedef MutableArray< RangeFieldType > DofStorageType;

private:
  enum { myId_ = 0};

public:
 //! Type of the underlying grid
  typedef typename DiscreteFunctionSpaceType::GridType GridType;

  typedef typename DiscreteFunctionSpaceType::Traits::MapperType MapperType;
  typedef typename DiscreteFunctionSpaceType::Traits::RangeFieldType DofType;
//! Type of the dof iterator used in the discrete function implementation.
  typedef typename DofStorageType::DofIteratorType DofIteratorType;
 //! Type of the constant dof iterator used in the discrete function implementation
  typedef typename DofStorageType::ConstDofIteratorType ConstDofIteratorType;

  //! Type of this class 
  typedef ProductDiscreteFunction <DiscreteFunctionSpaceType,DiscreteFunctionSpace2Type> DiscreteFunctionType;
  typedef DiscreteFunctionType ThisType;
  
  //! Type of class AdaptiveDiscreteFunction
  typedef AdaptiveDiscreteFunction<DiscreteFunctionSpaceType> DiscreteFunction1Type;
  typedef typename DiscreteFunction1Type::DofStorageType DofStorage1Type;

  //! Type of the discrete function implementation
  typedef DiscreteFunctionSpaceType FunctionSpaceType;
  typedef ProductDiscreteFunctionTraits<DiscreteFunctionSpaceType> Traits;

  /** \brief For ISTL-compatibility */
  typedef FieldVector<DofType,1> block_type;
  
public:

 /** \brief Constructor storing discrete function spaces 
        \param[in] f discrete function space  
	\param[in] f2 discrete function space2
 */
  ProductDiscreteFunction(const DiscreteFunctionSpaceType& f, const DiscreteFunctionSpace2Type& f2) ;
  
  //! delete stack of free local functions belonging to this discrete
  //! function 
  virtual ~ProductDiscreteFunction ();
       
  // ***********  Interface  *************************
  
  /** \brief returns local function \f$  u_{j_0}(x) = \sum _{i=1} ^{n_x}  u\left[j_0 n_x +i\right]\varphi _i(x) \f$ for global dof index \f$j_0\f$ of second space
      \param[in] dofIndex2 global dof index of second space
      \returns \f$  u_{j_0}(x) = \sum _{i=1} ^{n_x}  u\left[j_0 n_x +i\right]\varphi _i(x) \f$
   */
  inline DiscreteFunction1Type
  localFunction(int dofIndex2) const; 
  
  /** \brief returns local function \f$  u_{j_0,en_2}(x) = \sum _{i=1} ^{n_x}  u\left[global(j_0,en_2) n_x +i\right]\varphi _i(x) \f$ for local dof index \f$j_0\f$ of second space and given entity2
      \param[in] en2 entity of second space
      \param[in] dofIndex2 global dof index of second space
      \returns \f$  u_{j_0,en_2}(x) = \sum _{i=1} ^{n_x}  u\left[global(j_0,en_2) n_x +i\right]\varphi _i(x) \f$
   */
  template <class Entity2Type>
  inline DiscreteFunction1Type
  localFunction(const Entity2Type &en2, int dofIndex2) const;
  
  /** \brief local function \f$  u_{loc_2,en_2}(x) = \sum _{i=1} ^{n_x} \left[ \sum_{k=1} ^{\#dof_2}  u\left[global(k,en_2) n_x +i\right]\psi _k (loc_2) \right]\varphi _i(x) \f$ for given entity2 and local coordinate of second space and discrete function of first space
      \param[in] en2 entity of second space
      \param[in] loc2 local coordinate of second space
      \param[out] discFunc \f$  = \sum _{i=1} ^{n_x} \left[ \sum_{k=1} ^{\#dof_2}  u\left[global(k,en_2) n_x +i\right]\psi _k (loc_2) \right]\varphi _i(x) \f$
   */
  template < class Entity2Type, class LocalCoord2Type>
  inline
  void localFunction(const Entity2Type &en2, const LocalCoord2Type &loc2,  DiscreteFunction1Type &discFunc) const;

  /** \brief local function \f$  u_{qP_2,en_2}(x) = \sum _{i=1} ^{n_x} \left[ \sum_{k=1} ^{\#dof_2}  u\left[global(k,en_2) n_x +i\right]\psi _k (qP_2) \right]\varphi _i(x) \f$ for given entity2, quadrature type and quadrature point number of second space and discrete function of first space
      \param[in] en2 entity of second space
      \param[in] quad2 quadrature of second space
      \param[in] pointNr number of quadrature point
      \param[out] discFunc \f$ = \sum _{i=1} ^{n_x} \left[ \sum_{k=1} ^{\#dof_2}  u\left[global(k,en_2) n_x +i\right]\psi _k (qP_2) \right]\varphi _i(x) \f$
   */
  template < class Entity2Type, class QuadratureType>
  inline
  void localFunction(const Entity2Type &en2, const QuadratureType &quad2, int pointNr, DiscreteFunction1Type &discFunc) const;

 
 /** \brief returns dof iterator pointing to the first degree of freedom of this discrete function 
     \return dof iterator pointing to first dof 
 */
  inline
  DofIteratorType dbegin ( );
  
 /** \brief returns dof iterator pointing behind the last degree of freedom of this discrete function 
     \return dof iterator pointing behind the last dof 
  */
  inline
  DofIteratorType dend   ( );

  /** \brief returns dof iterator pointing to the first degree of freedom of this discrete function 
      \return dof iterator pointing to first dof 
   */
   inline
  ConstDofIteratorType dbegin ( ) const;
  
  /** \brief returns dof iterator pointing behind the last degree of freedom of this discrete function 
      \return dof iterator pointing behind the last dof 
   */  
  inline
  ConstDofIteratorType dend   ( ) const;

   /** \brief set all degrees of freedom to zero
    */  
  inline
  void clear( );

   /** \brief axpy operation
       \param[in] g discrete function that is added 
       \param[in] c scalar value to scale 
    */  
  inline
  void axpy (const ThisType & g, const RangeFieldType &c); 
      
   /** \brief print all degrees of freedom of this function to stream (for debugging purpose)
       \param[out] s std::ostream (e.g. std::cout)
    */
  inline
  void print(std::ostream& s) const; 

  /** \brief returns name of discrete function 
      \return string holding name of discrete function 
  */ 
  std::string name () const { return name_; }

  /** \brief returns total number of degrees of freedom, i.e. size of discrete function space 
      \return total number of dofs 
  */ 
  int size() const { return dofVec_.size(); }
  
  //! return pointer to internal array for use of BLAS routines 
  DofType * leakPointer () { return dofVec_.leakPointer();  };
  //! return pointer to internal array for use of BLAS routines 
  const DofType * leakPointer () const { return dofVec_.leakPointer(); };
  
  //! Get access to the related function space
  const DiscreteFunctionSpaceType& space() const { return functionSpace_; }
  //! Get access to the related function space2
  const DiscreteFunctionSpace2Type& space2() const { return functionSpace2_; }


private:  
  // name of this func
  std::string name_;

  //! The related function space
  const DiscreteFunctionSpaceType & functionSpace_;
  //! The related function space2
  const DiscreteFunctionSpace2Type & functionSpace2_;

  //! array containing the dof of this function, see dofmanager.hh
  //! the array is stored within the mem object 
  mutable DofStorageType dofVec_;

}; // end class ProductDiscreteFunction 


} // end namespace Fem

} // end namespace Dune

#include "productfunctionimp.cc"
#endif
