/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_GRIDPART_TEST_CHECKGEOMETRY_HH
#define DUNE_FEM_GRIDPART_TEST_CHECKGEOMETRY_HH

//- dune-common includes
#include <dune/common/forloop.hh>
#include <dune/common/exceptions.hh>
#include <dune/common/stdstreams.hh>

//- dune-grid includes
#include <dune/grid/test/checkgeometry.cc>

//- dune-fem includes
#include <dune/fem/gridpart/common/capabilities.hh>
#include <dune/fem/gridpart/test/failure.hh>


/** \brief Check geometries for all entities in the
 *         grid part. The tests are implemented in
 *         \code
<dune/grid/test/checkgeometry.cc>         
 *         \endcode
 *         Thus, failure handlers passed to this test
 *         will be ignored.
 *
 */

namespace Dune
{
  namespace Fem
  {

    template< class GridPartType, class FailureHandler >
    class CheckGeometry
    {
      template< class GridPart, bool >
      struct If
      {
        template< int codim, class Entity  >
        static void check ( const Entity&, FailureHandler & )
        { }
      };

      template< class GridPart >
      struct If< GridPart, true >
      {
        template< int codim, class Entity  >
        static void check ( const Entity &entity, FailureHandler & )
        {
          for ( int i = 0; i < entity.template count< codim >(); ++i )
          {
            typedef typename Entity::template Codim< codim >::EntityPointer SubEP;
            const SubEP subEP = entity.template subEntity< codim >( i );
            const typename SubEP::Entity &subEn = *subEP;
            const typename SubEP::Entity::Geometry &subGeo = subEn.geometry();

            if( subEn.type() != subGeo.type() )
              std::cerr << "Error: Entity and geometry report different geometry types on codimension " << codim << "." << std::endl;
            Dune::checkGeometry( subGeo );
          }
        }
      };

      template< int codim >
      struct CheckSubEntityGeometry
      {
        template< class Entity >
        static void apply ( const Entity &entity, FailureHandler &failureHandler )
        {
          // check whether grid part has entity of given codimension
          const bool hasEntity = Dune::Fem::GridPartCapabilities::hasEntity< GridPartType, codim >::v;
          If< GridPartType, hasEntity >::template check< codim, Entity >( entity, failureHandler );
        }
      };

    public:
      static const int dimension = GridPartType::dimension;

      static void check ( const GridPartType &gridPart, FailureHandler &failureHandler )
      {
        // tests will be forwarded to dune-grid, where failures are not supported
        std::cout << "Note: FailureHandler will be ignored" << std::endl;

        typedef typename GridPartType::template Codim< 0 >::IteratorType IteratorType;
        typedef typename GridPartType::template Codim< 0 >::EntityType EntityType;

        const IteratorType end = gridPart.template end< 0>();
        for( IteratorType it = gridPart.template begin< 0 >(); it != end; ++it )
        {
          const EntityType &entity = *it;
          ForLoop< CheckSubEntityGeometry, 0, dimension >::apply( entity, failureHandler );
        }
      }
    };

  } // end namespace Fem

} // end namespace Dune

#endif // #ifndef DUNE_FEM_GRIDPART_TEST_CHECKGEOMETRY_HH
