/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_MPIMANAGER_HH
#define DUNE_FEM_MPIMANAGER_HH

#include <dune/common/mpicollectivecommunication.hh>
#include <dune/common/mpihelper.hh>

namespace Dune
{

  struct MPIManager
  {
    typedef Dune::CollectiveCommunication< MPIHelper::MPICommunicator >
      CollectiveCommunication;

  private:
    MPIManager ()
    : helper_( 0 ),
      comm_( 0 )
    {}

    ~MPIManager ()
    {
      delete comm_;
    }

    // prohibit copying and assignment
    MPIManager ( const MPIManager & );
    MPIManager &operator= ( const MPIManager & );
    
    static MPIManager &instance ()
    {
      static MPIManager instance;
      return instance;
    }

  public:
    static void initialize ( int &argc, char **&argv )
    {
      MPIHelper *&helper = instance().helper_;
      CollectiveCommunication *&comm = instance().comm_;
#if HAVE_MPI && MPI_2 
#ifdef USE_SMP_PARALLEL 
      int provided;
      // use MPI_Init_thread for hybrid parallel programs 
      int is_initialized = MPI_Init_thread(&argc, &argv, MPI_THREAD_SINGLE, &provided );

      if( is_initialized != MPI_SUCCESS ) 
        DUNE_THROW(InvalidStateException,"MPI_Init_thread failed!");

      if( provided != MPI_THREAD_SINGLE ) 
        DUNE_THROW(InvalidStateException,"Wrong MPI mode initialized " << provided );
      /*
#if not defined NDEBUG && defined DUNE_DEVEL_MODE
      if( provided != MPI_THREAD_FUNNELED )
      {
        if( provided == MPI_THREAD_SINGLE )
          dwarn << "MPI thread support = single (instead of funneled)!" << std::endl;
        else
          dwarn << "WARNING: MPI thread support = " << provided << " != MPI_THREAD_FUNNELED " << MPI_THREAD_FUNNELED << std::endl;
      }
#endif // end NDEBUG 
      */

#endif // end USE_SMP_PARALLEL
#endif // end HAVE_MPI && MPI_2       

      if( (helper != 0) || (comm != 0) )
        DUNE_THROW( InvalidStateException, "MPIManager has already been initialized." );

      helper = &MPIHelper::instance( argc, argv );
      comm = new CollectiveCommunication( helper->getCommunicator() );
    }

    static MPIHelper &helper () DUNE_DEPRECATED
    {
      MPIHelper *const helper = instance().helper_;
      if( helper == 0 )
        DUNE_THROW( InvalidStateException, "MPIManager has not been initialized." );
      return *helper;
    }

    static const CollectiveCommunication &comm ()
    {
      const CollectiveCommunication *const comm = instance().comm_;
      if( comm == 0 )
        DUNE_THROW( InvalidStateException, "MPIManager has not been initialized." );
      return *comm;
    }

    static int rank ()
    {
      return comm().rank();
    }

    static int size ()
    {
      return comm().size();
    }

  private:
    MPIHelper *helper_;
    CollectiveCommunication *comm_;
  };

}

#endif // #ifndef DUNE_FEM_MPIMANAGER_HH
