/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_BOUNDARYMODEL_HH
#define DUNE_FEM_BOUNDARYMODEL_HH

#include <dune/common/bartonnackmanifcheck.hh>

namespace Dune
{

  namespace Fem
  {

  class DefaultBoundaryModelProperties
  {
  public:
    enum { hasDirichletValues = true };
    enum { hasNeumannValues = true };
    enum { hasRobinValues = true };
  };



  template< class FunctionSpaceImp,
            class BoundaryModelImp,
            class PropertiesImp >
  class BoundaryModelInterface
  {
  public:
    typedef FunctionSpaceImp FunctionSpaceType;
    typedef PropertiesImp Properties;

  private:
    typedef BoundaryModelInterface< FunctionSpaceType, BoundaryModelImp, Properties >
      ThisType;

  public:
    typedef ThisType BoundaryModelInterfaceType;

    enum BoundaryType { Dirichlet, Neumann, Robin };

    typedef typename FunctionSpaceType :: DomainType DomainType;
    typedef typename FunctionSpaceType :: RangeType RangeType;
    typedef typename FunctionSpaceType :: JacobianRangeType JacobianRangeType;

    typedef typename FunctionSpaceType :: DomainFieldType DomainFieldType;
    typedef typename FunctionSpaceType :: RangeFieldType RangeFieldType;

  public:
    /** \brief obtain type of boundary intersection
     *
     *  \param[in] intersection intersection whose boundary type shall be returned
     *
     *  \returns boundary type of the intersection
     */
    template< class IntersectionType >
    inline BoundaryType boundaryType ( const IntersectionType &intersection ) const
    {
      CHECK_INTERFACE_IMPLEMENTATION
        ( asImp().boundaryType( intersection ) );
      return asImp().boundaryType( intersection );
    }

    /** \brief evaluate Dirichlet boundary value
     *
     *  Given an intersection and a quadrature-point pair, evaluate the
     *  Dirichlet boundary data in the specified point.
     *
     *  \note The entity the boundary belongs to can be obtained from the
     *        intersection by the intersecion.inside method.
     *
     *  \note The quadrature is not guaranteed to be a codim-1-quadrature
     *
     *  \param[in]  intersection intersection on which the quadrature point lies
     *
     *  \param[in]  quadrature   quadrature to use
     *
     *  \param[in]  point        index of the evaluation point within the
     *                           quadrature
     *
     *  \param[out] phi          Dirichlet value in the quadrature point
     */
    template< class IntersectionType, class QuadratureType >
    inline void dirichletValues ( const IntersectionType &intersection,
                                  const QuadratureType &quadrature,
                                  int point,
                                  RangeType &phi ) const
    {
      assert( Properties :: hasDirichletValues );
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION
        ( asImp().dirichletValues( intersection, quadrature, point, phi ) );
    }

    template< class IntersectionType, class QuadratureType >
    inline void neumannValues ( const IntersectionType &intersection,
                                const QuadratureType &quadrature,
                                int point,
                                RangeType &phi ) const
    {
      assert( Properties :: hasNeumannValues );
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION
        ( asImp().neumannValues( intersection, quadrature, point, phi ) );
    }
    
    template< class IntersectionType, class QuadratureType >
    inline void robinValues ( const IntersectionType &intersection,
                              const QuadratureType &quadrature,
                              int point,
                              RangeType &phi ) const
    {
      assert( Properties :: hasRobinValues );
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION
        ( asImp().robinValues( intersection, quadrature, point, phi ) );
    }

    template< class IntersectionType, class QuadratureType >
    inline RangeFieldType robinAlpha ( const IntersectionType &intersection,
                                       const QuadratureType &quadrature,
                                       int point ) const
    {
      assert( Properties :: hasRobinValues );
      CHECK_INTERFACE_IMPLEMENTATION
        ( asImp().robinAlpha( intersection, quadrature, point ) );
      return asImp().robinAlpha( intersection, quadrature, point );
    }

  protected:
    inline const BoundaryModelImp &asImp () const
    {
      return static_cast< const BoundaryModelImp& >( *this );
    }

    inline BoundaryModelImp &asImp ()
    {
      return static_cast< BoundaryModelImp& >( *this );
    }
  };



  template< class FunctionSpaceImp,
            class BoundaryModelImp,
            class PropertiesImp = DefaultBoundaryModelProperties >
  class BoundaryModelDefault
  : public BoundaryModelInterface< FunctionSpaceImp, BoundaryModelImp, PropertiesImp >
  {
  public:
    typedef FunctionSpaceImp FunctionSpaceType;
    typedef PropertiesImp Properties;

  private:
    typedef BoundaryModelDefault< FunctionSpaceType, BoundaryModelImp, Properties >
      ThisType;
    typedef BoundaryModelInterface< FunctionSpaceType, BoundaryModelImp, Properties >
      BaseType;

  public:
    typedef typename FunctionSpaceType :: DomainType DomainType;
    typedef typename FunctionSpaceType :: RangeType RangeType;
    typedef typename FunctionSpaceType :: JacobianRangeType JacobianRangeType;

    typedef typename FunctionSpaceType :: DomainFieldType DomainFieldType;
    typedef typename FunctionSpaceType :: RangeFieldType RangeFieldType;
    
  public:
    template< class IntersectionType, class QuadratureType >
    inline void dirichletValues ( const IntersectionType &intersection,
                                  const QuadratureType &quadrature,
                                  int point,
                                  RangeType &phi ) const
    {
      assert( Properties :: hasDirichletValues );
      phi = 0;
    }

    template< class IntersectionType, class QuadratureType >
    inline void neumannValues ( const IntersectionType &intersection,
                                const QuadratureType &quadrature,
                                int point,
                                RangeType &phi ) const
    {
      assert( Properties :: hasNeumannValues );
      phi = 0;
    }
    
    template< class IntersectionType, class QuadratureType >
    inline void robinValues ( const IntersectionType &intersection,
                              const QuadratureType &quadrature,
                              int point,
                              RangeType &phi ) const
    {
      assert( Properties :: hasRobinValues );
      phi = 0;
    }

    template< class IntersectionType, class QuadratureType >
    inline RangeFieldType robinAlpha ( IntersectionType &intersection,
                                       QuadratureType &quadrature,
                                       int point ) const
    {
      assert( Properties :: hasRobinValues );
      return 0;
    }
  };

  } // end namespace Fem

  // #if DUNE_FEM_COMPATIBILITY  
  // put this in next version 1.4 

  using Fem :: BoundaryModelDefault ;

  // #endif // DUNE_FEM_COMPATIBILITY


} // end namespace Dune

#endif
