/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_ELLIPTICDISCRETEMODEL_HH
#define DUNE_ELLIPTICDISCRETEMODEL_HH


//- Dune includes 
#include <dune/common/fvector.hh>
#include <dune/common/bartonnackmanifcheck.hh>

//- local includes 
#include <dune/fem/misc/boundaryidentifier.hh>
#include "selection.hh"
#include "dgdiscretemodel.hh"

namespace Dune
{

  /**
   * @brief Interface class for problem definition in the 
   * DG primal formulation context.
   *
   */
  template <class DiscreteModelTraits>
  class EllipticDiscreteModelInterface 
  {
  public:
    //! Traits class defined by the user
    typedef DiscreteModelTraits Traits;
    //! Implementation type for Barton-Nackman trick
    typedef typename Traits::DiscreteModelType DiscreteModelType;
    //! Function space (we take the discrete one even though the continuous one
    //! would suffice...)
    typedef typename Traits::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;
    //! Coordinate type (world coordinates)
    typedef typename DiscreteFunctionSpaceType::DomainType DomainType;
    //! Vector type of the function space's range
    typedef typename DiscreteFunctionSpaceType::RangeType RangeType;
    //! Matrix type of the function space's jacobian
    typedef typename DiscreteFunctionSpaceType::JacobianRangeType JacobianRangeType;
    //! type of range field 
    typedef typename DiscreteFunctionSpaceType::RangeFieldType RangeFieldType;

    //! Type of GridPart 
    typedef typename DiscreteFunctionSpaceType::GridPartType GridPartType; 
    //! Type of the grid
    typedef typename GridPartType::GridType GridType;

    //! dimension of grid 
    enum { dim = GridType :: dimension };

    //! type of domain for faces 
    typedef FieldVector<RangeFieldType,dim-1> FaceDomainType; 

    //! Intersection iterator of the grid
    typedef typename GridPartType::IntersectionIteratorType IntersectionIteratorType;
    //! Intersection of the grid
    typedef typename IntersectionIteratorType::IntersectionType IntersectionType;
    //! Element (codim 0 entity) of the grid
    typedef typename GridType::template Codim<0>::Entity EntityType;

    //! type of boundary identifier 
    typedef BoundaryIdentifier BoundaryIdentifierType; 

  public:
    //! \todo Please document
    bool constantCoefficient () const
    {
      CHECK_INTERFACE_IMPLEMENTATION( asImp().constantCoefficient() );
      return asImp().constantCoefficient();
    }

    //! Returns true if problem has a coefficient before laplace term.
    bool hasCoefficient () const 
    { 
      CHECK_INTERFACE_IMPLEMENTATION(asImp().hasCoefficientFlux());
      return asImp().hasCoefficientFlux(); 
    }

    //! Returns true if problem has right hand side not qual to zero 
    bool hasRHS () const 
    { 
      CHECK_INTERFACE_IMPLEMENTATION(asImp().hasRHS());
      return asImp().hasRHS(); 
    }

    //! \brief Computes the flux at the boundary
    //! Special kind of numerical flux. The intersection iterator provides
    //! the necessary information to identify the type of boundary.
    //! \param it Iterator of the face in consideration.
    //! \param time Global time.
    //! \param local local point where the boundaryvalue is evaluated
    //! \param u Tuple of the states of the previous passes and of the global argument.
    //! \param bndVal boundary Value (return value)
    template <class FaceDomType,
              class ArgumentTuple>
    BoundaryIdentifierType
    DUNE_DEPRECATED
    boundaryValue(const IntersectionIteratorType& it,
                  double time, const FaceDomType& local,
                  const ArgumentTuple& u,
                  RangeType& bndVal) const
    {
      CHECK_INTERFACE_IMPLEMENTATION(asImp().boundaryValue(it,time,local,u,bndVal));
      return asImp().boundaryValue(it,time,local,u,bndVal);
    }

    // Ensure: entities set correctly before call
    template <class FaceDomType,
              class ArgumentTuple>
    BoundaryIdentifierType
    boundaryValue(const IntersectionType& it,
                  double time, const FaceDomType& local,
                  const ArgumentTuple& u,
                  RangeType& bndVal) const
    {
      CHECK_INTERFACE_IMPLEMENTATION(asImp().boundaryValue(it,time,local,u,bndVal));
      return asImp().boundaryValue(it,time,local,u,bndVal);
    }

    template <class ArgumentTuple, class CoefficientType>
    DUNE_DEPRECATED
    void coefficientFace(const IntersectionIteratorType& it,
                         const double time,
                         const FaceDomainType& local,
                         const ArgumentTuple& uLeft,
                         const ArgumentTuple& uRight,
                         CoefficientType & coeffLeft,
                         CoefficientType & coeffRight) const
    {
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION(asImp().coefficientFace(it,time,local,
                        uLeft,uRight,coeffLeft,coeffRight));
    }

    // Ensure: entities set correctly before call
    template <class ArgumentTuple, class CoefficientType>
    void coefficientFace(const IntersectionType& it,
                         const double time,
                         const FaceDomainType& local,
                         const ArgumentTuple& uLeft,
                         const ArgumentTuple& uRight,
                         CoefficientType & coeffLeft,
                         CoefficientType & coeffRight) const
    {
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION(asImp().coefficientFace(it,time,local,
                        uLeft,uRight,coeffLeft,coeffRight));
    }
 
    //! \brief Implements the source term of the problem.
    //! Source term contribution. The source term can contain parts of the
    //! non-conservative contributions.
    //! \param en Entity where the flux is to be evaluated
    //! \param time Global time.
    //! \param local Reference element coordinate where the flux is evaluated
    //! \param u Tuple of the states of the previous passes and of the global
    //! argument.
    //! \param jac Tuple of the gradients of the precedings passes' results 
    //! (needed for the non-conservative contributions)
    //! \param s The source contribution (return value).
    template <class ArgumentTuple, class JacobianTuple>
    void source(EntityType& en, 
                double time, const DomainType& local,
                const ArgumentTuple& u, 
                const JacobianTuple& jac, 
                RangeType& s) 
    { 
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION(
        asImp().source(en,time,local,u,jac,s));
    }

    //! \brief Implements the right hand side of the problem.
    //! Right hand side term contribution.
    //! \param en Entity where the flux is to be evaluated
    //! \param time Global time.
    //! \param local Reference element coordinate where the flux is evaluated
    //! \param u Tuple of the states of the previous passes and of the global
    //! argument.
    //! \param jac Tuple of the gradients of the precedings passes' results 
    //! (needed for the non-conservative contributions)
    //! \param rhs The right hand side contribution (return value).
    template <class ArgumentTuple, class JacobianTuple>
    void rightHandSide(const EntityType &en,
                       const double time,
                       const DomainType& local,
                       const ArgumentTuple& u,
                       const JacobianTuple& jac,
                       RangeType& rhs)
    {
      CHECK_AND_CALL_INTERFACE_IMPLEMENTATION(
        asImp().rightHandSide(en,time,local,u,jac,rhs));
    }

  protected:
    // don't create instances from this class 
    EllipticDiscreteModelInterface() {}

  private:
    DiscreteModelType& asImp() { return static_cast<DiscreteModelType&>(*this); }
    const DiscreteModelType& asImp() const { return static_cast<const DiscreteModelType&>(*this); }
  };
  
}  // end namespace Dune

#undef CHECK_INTERFACE_IMPLEMENTATION
#undef CHECK_AND_CALL_INTERFACE_IMPLEMENTATION

#endif
