/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <cassert>
#if HAVE_DUNE_GEOMETRY
#include <dune/geometry/referenceelements.hh>
#else
#include <dune/grid/common/genericreferenceelements.hh>
#endif

#include <dune/fem/space/basefunctions/storageinterface.hh>
#include <dune/fem/misc/threadmanager.hh>

namespace Dune {

  namespace Fem {

  template <class ct, int dim>
  typename PointProvider<ct, dim, 0>::PointContainerType
  PointProvider<ct, dim, 0>::points_;

  template <class ct, int dim>
  void PointProvider<ct, dim, 0>::
  registerQuadrature(const QuadratureType& quad)
  {
    QuadratureKeyType key( quad.geometryType(), quad.id() );
    
    if (points_.find( key ) == points_.end() ) 
    {
      // only register when in single thread mode 
      assert( Fem :: ThreadManager :: singleThreadMode() );

      PointIteratorType it =
        points_.insert(std::make_pair
                       (key,
                        GlobalPointVectorType(quad.nop()))
                       ).first;
      for (size_t i = 0; i < quad.nop(); ++i) 
      {
        it->second[i] = quad.point(i);
      }

      // register quadrature to existing base function storages 
      StorageInterface<dim>::registerQuadratureToStorages(quad);
    }
  }

  template <class ct, int dim>
  const typename PointProvider<ct, dim, 0>::GlobalPointVectorType&
  PointProvider<ct, dim, 0>::getPoints(const size_t id, const GeometryType& elementGeo) 
  {
    QuadratureKeyType key( elementGeo, id );

    typename PointContainerType::const_iterator pos = points_.find( key );
#ifndef NDEBUG
    if( pos == points_.end() )
    {
      std::cerr << "Unable to find quadrature points in list (key = " << key << ")." << std::endl;
      for( pos = points_.begin(); pos != points_.end(); ++pos )
        std::cerr << "found key: " << pos->first << std::endl;
      std::cerr << "Aborting..." << std::endl;
      abort();
    }
#endif
    return pos->second;
  }

  template <class ct, int dim>
  typename PointProvider<ct, dim, 1>::PointContainerType
  PointProvider<ct, dim, 1>::points_;

  template <class ct, int dim>
  typename PointProvider<ct, dim, 1>::MapperContainerType
  PointProvider<ct, dim, 1>::mappers_;

  template <class ct, int dim>
  const typename PointProvider<ct, dim, 1>::MapperVectorType&
  PointProvider<ct, dim, 1>::getMappers(const QuadratureType& quad,
                                        const GeometryType& elementGeo) 
  {
    QuadratureKeyType key( elementGeo , quad.id() );
    
    MapperIteratorType it = mappers_.find( key );
    if (it == mappers_.end()) {
      std::vector<LocalPointType> pts(quad.nop());
      for (size_t i = 0; i < quad.nop(); ++i) {
        pts[i] = quad.point(i);
      }
      it = addEntry(quad, pts, elementGeo);
    }

    return it->second;
  }

  template <class ct, int dim>
  const typename PointProvider<ct, dim, 1>::MapperVectorType&
  PointProvider<ct, dim, 1>::getMappers(const QuadratureType& quad,
                                        const LocalPointVectorType& pts,
                                        const GeometryType& elementGeo)
  {
    QuadratureKeyType key( elementGeo, quad.id() );
    
    MapperIteratorType it = mappers_.find( key );
    if (it == mappers_.end()) {
      it = addEntry(quad, pts, elementGeo);
    }

    return it->second;
  }

  template <class ct, int dim>
  const typename PointProvider<ct, dim, 1>::GlobalPointVectorType&
  PointProvider<ct, dim, 1>::getPoints(const size_t id, const GeometryType& elementGeo)
  {
    QuadratureKeyType key( elementGeo, id );
    
    assert(points_.find(key) != points_.end());
    return points_.find(key)->second;
  }

  template <class ct, int dim>
  typename PointProvider<ct, dim, 1>::MapperIteratorType
  PointProvider<ct, dim, 1>::addEntry(const QuadratureType& quad,
                                      const LocalPointVectorType& points,
                                      GeometryType elementGeo)
  {
    // amke sure we are in single thread mode 
    assert( Fem :: ThreadManager::singleThreadMode() );

    // generate key 
    QuadratureKeyType key ( elementGeo, quad.id() );
      

    const GenericReferenceElement<ct, dim>& refElem = 
      GenericReferenceElements<ct, dim>::general(elementGeo);

    const int numLocalPoints = points.size();
    const int numFaces = refElem.size(codim);
    const int numGlobalPoints = numFaces*numLocalPoints;
    
    PointIteratorType pit = 
      points_.insert(std::make_pair(key, 
                                    GlobalPointVectorType(numGlobalPoints))).first;
    MapperIteratorType mit =
      mappers_.insert(std::make_pair(key,
                                     MapperVectorType(numFaces))).first;
    int globalNum = 0;
    for (int face = 0; face < numFaces; ++face) 
    {
      // Assumption: all faces have the same type
      // (not true for pyramids and prisms)
      //assert(sameGeometry(quad.geometryType(), refElem.type(face, codim)));
      MapperType pMap(numLocalPoints);
        
      for (int pt = 0; pt < numLocalPoints; ++pt, ++globalNum) {
        // Store point on reference element
        pit->second[globalNum] = 
          refElem.template global<codim>(points[pt], face, codim);
        
        // Store point mapping
        pMap[pt] = globalNum;
      }
      mit->second[face] = pMap;  // = face*numLocalPoints+pt
    } // end for all faces

    // register quadrature to base function storages 
    StorageInterface<dim>::registerQuadratureToStorages(quad, elementGeo, 1);
    
    return mit;
  }

  template <class ct, int dim>
  bool PointProvider<ct, dim, 1>::sameGeometry(GeometryType geo1, 
                                               GeometryType geo2)
  {
    return geo1 == geo2;
  }  
  } // end namespace Fem
} // end namespace Dune
