/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <config.h>

#include <dune/grid/sgrid.hh>

#ifdef ENABLE_ALUGRID
#include <dune/grid/alugrid.hh>
#endif

#ifdef ENABLE_ALBERTA
#include <dune/grid/albertagrid.hh>
#endif

#ifdef ENABLE_UGGRID
#define USE_PRISMGRID
#include <dune/grid/uggrid.hh>
#endif

#ifdef USE_PRISMGRID
//#include <dune/prismgrid/dgfgridtype.hh> 
#endif

#include <dune/geometry/referenceelements.hh>

#include "cache_test.hh"

namespace Dune {
  namespace Fem {
  
  void CacheProvider_Test::run() 
  {
    hexaTest();
    tetraTest();
    prismTest();
    triangleTest();
    quadTest();
  }

  void CacheProvider_Test::prismTest()
  {
#ifdef USE_PRISMGRID 
    const int dim = 3;
    const int codim = 1;

    //typedef PrismGrid<dim, dim> GridType;
    typedef UGGrid<dim> GridType;
    
    typedef CacheProvider<GridType, codim> CacheProviderType;
    typedef Quadrature< GridType :: ctype, GridType :: dimension-1> QuadratureType;

    typedef CacheProviderType::MapperType MapperType;
    typedef PointProvider<double, dim, codim> PointProviderType;
    typedef PointProviderType::GlobalPointVectorType PointVectorType;

    GeometryType elemGeo = GeometryType(GeometryType::prism,3);

    // Get reference element
    const GenericReferenceElement<double, dim>& refElem =
      GenericReferenceElements<double, dim>::general(elemGeo);
    
    // Loop over all faces
    for (int i = 0; i < refElem.size(codim); ++i) 
    {
      GeometryType faceGeo = (refElem.size(i,codim,dim) == dim) ? 
          GeometryType(GeometryType::simplex,2) :
          GeometryType(GeometryType::cube,2);

      // Build quadrature
      QuadratureType quad(faceGeo, 3);

      // Ask for one mapper so that the points get registered
      CacheProviderType::getMapper(quad, elemGeo, 0, 0);

      const PointVectorType& points =
        PointProviderType::getPoints(quad.id(), elemGeo);

      const MapperType& m = 
        CacheProviderType::getMapper(quad, elemGeo, i, 0);
      
      _test(m.size() == (size_t) quad.nop());
      
      // Loop over all points
      for (size_t j = 0; j < m.size(); ++j) 
      {
        for (int d = 0; d < dim; ++d) {
          _floatTest(points[m[j]][d], 
                     refElem.global<codim>(quad.point(j), i, codim)[d]);
        }
      }

    }
#endif
  }

  void CacheProvider_Test::hexaTest()
  {
#ifdef ENABLE_ALUGRID
    const int dim = 3;
    const int codim = 1;

    typedef ALUCubeGrid< 3, 3 > GridType;
    typedef CacheProvider<GridType, codim> CacheProviderType;
    typedef Quadrature< GridType :: ctype, GridType :: dimension-1> QuadratureType;
    //typedef CacheProviderType::QuadratureType QuadratureType;
    typedef CacheProviderType::MapperType MapperType;
    typedef PointProvider<double, dim, codim> PointProviderType;
    typedef PointProviderType::GlobalPointVectorType PointVectorType;

    GeometryType elemGeo = GeometryType(GeometryType::cube,3);
    GeometryType faceGeo = GeometryType(GeometryType::cube,2);

    // Get reference element
    const GenericReferenceElement<double, dim>& refElem =
      GenericReferenceElements<double, dim>::general(elemGeo);
    
    // Build quadrature
    QuadratureType quad(faceGeo, 3);

    // Ask for one mapper so that the points get registered
    CacheProviderType::getMapper(quad, elemGeo, 0, 0);

    const PointVectorType& points =
      PointProviderType::getPoints(quad.id(), elemGeo);

    // Loop over all faces
    for (int i = 0; i < refElem.size(codim); ++i) {
      const MapperType& m = 
        CacheProviderType::getMapper(quad, elemGeo, i, 0);
      
      _test(m.size() == (size_t) quad.nop());
      // Loop over all points
      for (size_t j = 0; j < m.size(); ++j) {
        for (int d = 0; d < dim; ++d) {
          _floatTest(points[m[j]][d], 
                     refElem.global<codim>(quad.point(j), i, codim)[d]);
        }
      }

    }
#endif
  }

  void CacheProvider_Test::tetraTest()
  {
#ifdef ENABLE_ALUGRID
    const int dim = 3;
    const int codim = 1;

    typedef ALUSimplexGrid< 3, 3 > GridType;
    typedef CacheProvider<GridType, codim> CacheProviderType;
    //typedef CacheProviderType::QuadratureType QuadratureType;
    typedef Quadrature< GridType :: ctype, GridType :: dimension-1> QuadratureType;
    typedef CacheProviderType::MapperType MapperType;
    typedef PointProvider<double, dim, codim> PointProviderType;
    typedef PointProviderType::GlobalPointVectorType PointVectorType;

    GeometryType elemGeo = GeometryType(GeometryType::simplex,3);
    GeometryType faceGeo = GeometryType(GeometryType::simplex,2);

    // Get reference element
    const GenericReferenceElement<double, dim>& refElem =
      GenericReferenceElements<double, dim>::general(elemGeo);
    
    // Build quadrature
    QuadratureType quad(faceGeo, 3);

    // Ask for one mapper so that the points get registered
    CacheProviderType::getMapper(quad, elemGeo, 0, 0);

    const PointVectorType& points =
      PointProviderType::getPoints(quad.id(), elemGeo);

    // Loop over all faces
    for (int i = 0; i < refElem.size(codim); ++i) {
      const MapperType& m = CacheProviderType::getMapper(quad, elemGeo, i, 0);
      
      _test(m.size() == (size_t) quad.nop());
      // Loop over all points
      for (size_t j = 0; j < m.size(); ++j) {
        for (int d = 0; d < dim; ++d) {
          _floatTest(points[m[j]][d], 
                     refElem.global<codim>(quad.point(j), i, codim)[d]);
        }
      }
    }
#endif
  }

  void CacheProvider_Test::triangleTest()
  {
#ifdef ENALBE_ALBERTA
    const int dim = 2;
    const int codim = 1;

    typedef AlbertaGrid< dim > GridType;
    typedef CacheProvider<GridType, codim> CacheProviderType;
    //typedef CacheProviderType::QuadratureType QuadratureType;
    typedef Quadrature< GridType :: ctype, GridType :: dimension-1> QuadratureType;
    typedef CacheProviderType::MapperType MapperType;
    typedef PointProvider<double, dim, codim> PointProviderType;
    typedef PointProviderType::GlobalPointVectorType PointVectorType;

    GeometryType elemGeo = GeometryType(GeometryType::simplex,2);
    GeometryType faceGeo = GeometryType(GeometryType::simplex,1);

    // Get reference element
    const GenericReferenceElement<double, dim>& refElem =
      GenericReferenceElements<double, dim>::general(elemGeo);
    
    // Build quadrature
    QuadratureType quad(faceGeo, 3);

    // Ask for one mapper so that the points get registered
    CacheProviderType::getMapper(quad, elemGeo, 0, 0);

    const PointVectorType& points =
      PointProviderType::getPoints(quad.id(), elemGeo);

    // Loop over all faces
    for (int i = 0; i < refElem.size(codim); ++i) {
      const MapperType& m = CacheProviderType::getMapper(quad, elemGeo, i, 0);
      
      _test(m.size() == (size_t) quad.nop());
      // Loop over all points
      for (size_t j = 0; j < m.size(); ++j) {
        for (int d = 0; d < dim; ++d) {
          _floatTest(points[m[j]][d], 
                     refElem.global<codim>(quad.point(j), i, codim)[d]);
        }
      }
    }
#endif
  }

  void CacheProvider_Test::quadTest()
  {
    const int dim = 2;
    const int codim = 1;

    typedef SGrid< dim, dim > GridType;
    typedef CacheProvider<GridType, codim> CacheProviderType;
    typedef Quadrature< GridType :: ctype, GridType :: dimension-1> QuadratureType;
    //typedef CacheProviderType::QuadratureType QuadratureType;
    typedef CacheProviderType::MapperType MapperType;
    typedef PointProvider<double, dim, codim> PointProviderType;
    typedef PointProviderType::GlobalPointVectorType PointVectorType;

    GeometryType elemGeo = GeometryType(GeometryType::cube,2);
    GeometryType faceGeo = GeometryType(GeometryType::cube,1);

    // Get reference element
    const GenericReferenceElement<double, dim>& refElem =
      GenericReferenceElements<double, dim>::general(elemGeo);
    
    // Build quadrature
    QuadratureType quad(faceGeo, 5);

    // Ask for one mapper so that the points get registered
    CacheProviderType::getMapper(quad, elemGeo, 0, 0);

    const PointVectorType& points =
      PointProviderType::getPoints(quad.id(), elemGeo);

    // Loop over all faces
    for (int i = 0; i < refElem.size(codim); ++i) {
      const MapperType& m = CacheProviderType::getMapper(quad, elemGeo, i, 0);
      
      _test(m.size() == (size_t) quad.nop());
      // Loop over all points
      for (size_t j = 0; j < m.size(); ++j) {
        for (int d = 0; d < dim; ++d) {
          _floatTest(points[m[j]][d], 
                     refElem.global<codim>(quad.point(j), i, codim)[d]);
        }
      }
    }
  }

  } // end namespace Fem
} // end namespace Dune
