/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_INTERSECTIONQUADRATURE_HH
#define DUNE_INTERSECTIONQUADRATURE_HH

//- Dune includes
#include <dune/common/misc.hh>

//- Local includes
#include "elementquadrature.hh"
#include "caching/twistutility.hh"
#include "caching/pointmapper.hh"
#include "caching/cacheprovider.hh"

#include "elementquadrature.hh"
#include "cachingquadrature.hh"

namespace Dune
{
  namespace Fem
  {  
  /** \brief IntersectionQuadrature is a helper class for creating the appropriate face quadratures 
             for integrating over intersections. */
  template< typename FaceQuadrature, bool conforming  >
  class IntersectionQuadrature 
  {
    template < typename FaceQuadratureImp, bool isConforming > 
    struct QuadSelector
    {
      // use given quadrature 
      typedef FaceQuadratureImp  FaceQuadratureType;
    };

    template < typename FaceQuadratureImp > 
    struct QuadSelector<FaceQuadratureImp, false>
    {
      // in this case non conforming type is used 
      typedef typename  FaceQuadratureImp :: 
        NonConformingQuadratureType  FaceQuadratureType;
    };

  public:
    //! type of grid partition
    typedef typename FaceQuadrature :: GridPartType GridPartType;
      
    //! type of the grid
    typedef typename GridPartType :: GridType GridType;

    //! Type of the intersection iterator
    typedef typename GridPartType::IntersectionIteratorType IntersectionIteratorType;
    typedef typename IntersectionIteratorType::Intersection IntersectionType;

    //! codimension of the element quadrature
    enum { codimension = FaceQuadrature :: codimension };
 
    //! type of intersection quadrature implementation 
    typedef typename QuadSelector<FaceQuadrature, conforming> :: FaceQuadratureType FaceQuadratureType; 
    
    //! Dimension of the world.
    enum { dimension = FaceQuadratureType ::dimension };

    //! Just another name for double...
    typedef typename FaceQuadratureType :: RealType RealType;
    //! The type of the coordinates in the codim-0 reference element.
    typedef typename FaceQuadratureType :: CoordinateType CoordinateType;

    typedef typename FaceQuadratureType::LocalCoordinateType LocalCoordinateType;

    //! for compatibility
    typedef typename GridType::template Codim< 0 >::Entity EntityType;
    
    /** \brief Constructor creating an inside and an outside face quadrature for
               integrating over an intersection. 
            
        \param[in]  gridPart      grid partition
        \param[in]  intersection  intersection
        \param[in]  order         desired order of the quadrature
 
     */ 
    IntersectionQuadrature( const GridPartType &gridPart, 
                            const IntersectionType &intersection,
                            const int order) 
    : inside_ ( gridPart, intersection, order, FaceQuadratureType :: INSIDE ),
      outside_( gridPart, intersection, order, FaceQuadratureType :: OUTSIDE ) 
    {}

    //! \brief return reference to inside face quadrature 
    const FaceQuadratureType& inside()  const { return inside_;  }

    //! \brief return reference to outside face quadrature 
    const FaceQuadratureType& outside() const { return outside_; }

    size_t nop () const
    {
      assert( inside().nop() == outside().nop() );
      return inside().nop();
    }

    const LocalCoordinateType &localPoint ( const int qp ) const
    {
      assert( inside().localPoint( qp ) == outside().localPoint( qp ) );
      return inside().localPoint( qp );
    }

    const RealType &weight ( const int qp ) const
    {
      assert( inside().weight( qp ) == outside().weight( qp ) );
      return inside().weight( qp );
    }

  private:
    // prohibit copying 
    IntersectionQuadrature( const IntersectionQuadrature & );

  protected:  
    const FaceQuadratureType inside_;
    const FaceQuadratureType outside_;
  };

  } // end namespace Fem

  // #if DUNE_FEM_COMPATIBILITY  
  // put this in next version 1.4 

  using Fem :: IntersectionQuadrature ;
  // #endif // DUNE_FEM_COMPATIBILITY

} // end namespace Dune 

#endif // #ifndef DUNE_INTERSECTIONQUADRATURE_HH
