/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_DEFAULTDATAHANDLE_HH
#define DUNE_FEM_DEFAULTDATAHANDLE_HH

#include <cassert>

//- Dune includes 
#include <dune/grid/common/datahandleif.hh>
#include <dune/fem/space/common/commoperations.hh>

namespace Dune
{

  namespace Fem 
  {

  /** \class   DefaultCommunicationHandler
   *  \ingroup DFComm
   *  \brief   Default communication handler for discrete functions
   *
   *  \param  DiscreteFunction  type of discrete function to be communicated
   */
  template< class DiscreteFunction, class Operation = DFCommunicationOperation::Add >
  class DefaultCommunicationHandler
  : public CommDataHandleIF
    < DefaultCommunicationHandler< DiscreteFunction, Operation >,
      typename DiscreteFunction::RangeFieldType >
  {
    typedef DefaultCommunicationHandler< DiscreteFunction, Operation > ThisType;
    typedef CommDataHandleIF< ThisType, typename DiscreteFunction::RangeFieldType > BaseType;

  public:  
    typedef typename BaseType::DataType DataType;

    typedef DiscreteFunction DiscreteFunctionType;

    typedef typename DiscreteFunctionType::DiscreteFunctionSpaceType DiscreteFunctionSpaceType;

  protected:
    typedef typename DiscreteFunctionSpaceType::BlockMapperType MapperType;

    typedef typename DiscreteFunctionType::DofBlockPtrType DofBlockPtrType;

    static const int blockSize = DiscreteFunctionSpaceType::localBlockSize;
    
  public:
    DefaultCommunicationHandler( DiscreteFunctionType &function )
    : function_( &function ),
      mapper_( function.space().blockMapper() )
    {} 
    
    DefaultCommunicationHandler( const DefaultCommunicationHandler &other )
    : function_( other.function_ ),
      mapper_( other.mapper_ )
    {}
    
  private:  
    //! cannot be implemented because of the reference
    DefaultCommunicationHandler &operator= ( const DefaultCommunicationHandler & );

  public:
    bool contains ( int dim, int codim ) const
    {
      return mapper_.contains( codim );
    }

    bool fixedsize ( int dim, int codim) const
    {
      return mapper_.fixedDataSize( codim );
    }

    //! read buffer and apply operation 
    template< class MessageBuffer, class Entity >
    void gather ( MessageBuffer &buffer, const Entity &entity ) const
    {
      const unsigned int numEntityDofs = mapper_.numEntityDofs( entity );
      for( unsigned int i = 0; i < numEntityDofs; ++i )
      {
        const unsigned int index = mapper_.mapEntityDofToGlobal( entity, i );
        
        DofBlockPtrType blockPtr = function_->block( index );
        for( int j = 0; j < blockSize; ++j )
          buffer.write( (*blockPtr)[ j ] );
      }
    }

    //! read buffer and apply operation 
    template< class MessageBuffer, class Entity >
    void scatter ( MessageBuffer &buffer, const Entity &entity, size_t n )
    {
      const unsigned int numEntityDofs = mapper_.numEntityDofs( entity );

      assert( n == blockSize * numEntityDofs );
      for( unsigned int i = 0; i < numEntityDofs; ++i )
      {
        const unsigned int index = mapper_.mapEntityDofToGlobal( entity, i );

        DofBlockPtrType blockPtr = function_->block( index );
        for( int j = 0; j < blockSize; ++j )
        {
          DataType value;
          buffer.read( value );

          Operation :: apply( value, (*blockPtr)[ j ] );
        }
      }
    }

    //! return local dof size to be communicated 
    template< class Entity >
    size_t size ( const Entity &entity ) const
    {
      return blockSize * mapper_.numEntityDofs( entity );
    }

  protected:
    DiscreteFunctionType *const function_;
    const MapperType &mapper_;
  };
  
  } // end namespace Fem 

} // end namespace Dune

#endif // #ifndef DUNE_FEM_DEFAULTDATAHANDLE_HH
