/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FVSPACE_INLINE_HH
#define DUNE_FVSPACE_INLINE_HH

//- system includes 
#include <algorithm>

//- Dune includes 
#include <dune/fem/space/common/allgeomtypes.hh>

#include <dune/fem/space/fvspace.hh>

namespace Dune {

namespace Fem {

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
FiniteVolumeSpace (GridPartType & gridPart,
        const InterfaceType commInterface,
        const CommunicationDirection commDirection)
: DefaultType(gridPart, commInterface, commDirection),
  baseFuncSet_(),
  blockMapper_( BlockMapperProviderType::getObject( gridPart ) ),
  mapper_( blockMapper_ )
{
  makeFunctionSpace(gridPart);
}

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline void FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
makeFunctionSpace (GridPartType& gridPart) 
{
  // search the macro grid for different element types 
  AllGeomTypes< typename GridPartType::IndexSetType ,
                typename GridPartType::GridType > 
            allGeomTypes(gridPart.indexSet());
  
  const std::vector<GeometryType>& geomTypes = allGeomTypes.geomTypes(0);
  std::size_t maxDofs = 0;

  for( std::size_t i = 0; i < geomTypes.size(); ++i )
  {
    const GeometryType & geoType = geomTypes[i];
    if(baseFuncSet_.find( geoType ) == baseFuncSet_.end() ) 
    {
      // get pointer to base function set 
      const BaseFunctionSetImp* baseSet = 
        & SingletonProviderType::getObject(geoType);
      
      baseFuncSet_[ geoType ] = baseSet; 
      maxDofs = std::max( maxDofs, baseSet->size() );
    }
  }
  assert( std::size_t( mapper().maxNumDofs() ) == maxDofs );
}
  
template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
~FiniteVolumeSpace () 
{
  typedef typename BaseFunctionMapType :: iterator iterator;
  iterator end = baseFuncSet_.end();
  for(iterator it = baseFuncSet_.begin(); it != end; ++it)
  {
    const BaseFunctionSetImp * set = (*it).second; 
    SingletonProviderType::removeObject(*set);
  }

  baseFuncSet_.clear();
  BlockMapperProviderType::removeObject(blockMapper_);
}  

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline DFSpaceIdentifier FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
type () const 
{
  return FiniteVolumeSpace_id;
}

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
template <class EntityType> 
inline const 
typename FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::BaseFunctionSetType  
FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
baseFunctionSet (const EntityType &en) const 
{
  return this->baseFunctionSet(en.type());
}

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline const 
typename FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::BaseFunctionSetType  
FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::
baseFunctionSet (const GeometryType& geomType) const 
{
  assert( baseFuncSet_.find(geomType) != baseFuncSet_.end());
  return BaseFunctionSetType(baseFuncSet_[geomType]);
}

template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline typename FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::MapperType&
FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::mapper() const {
  return mapper_;
}
template <class FunctionSpaceImp, class GridPartImp, int polOrd, template <class> class BaseFunctionStorageImp >
inline typename FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::BlockMapperType&
FiniteVolumeSpace<FunctionSpaceImp, GridPartImp, polOrd, BaseFunctionStorageImp>::blockMapper() const 
{
  return blockMapper_;
}
   
} // end namespace Fem  

} // end namespace Dune 
#endif
