/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FVSPACEBASEFUNCTIONS_HH
#define DUNE_FVSPACEBASEFUNCTIONS_HH

//- Dune includes 
#include <dune/common/misc.hh>
#include <dune/grid/common/grid.hh>

//- Dune-Fem includes 
#include <dune/fem/space/basefunctions/basefunctioninterface.hh>
#include <dune/fem/space/basefunctions/basefunctionfactory.hh>

namespace Dune
{

  namespace Fem 
  {

  //! definition of FVBaseFunction, implementation via specialization 
  template< class FunctionSpaceType, int polOrd >
  class FVBaseFunction;
           
  //! Piecewise const base functions for all types of elements 
  template< class FunctionSpaceType >
  class FVBaseFunction< FunctionSpaceType, 0 >
  : public BaseFunctionInterface< FunctionSpaceType >
  {
    typedef BaseFunctionInterface< FunctionSpaceType > BaseType;

    enum { dimRange = FunctionSpaceType::dimRange };
    int baseNum_;

    typedef typename FunctionSpaceType::DomainType DomainType;
    typedef typename FunctionSpaceType::RangeType RangeType;
    typedef typename FunctionSpaceType::RangeFieldType RangeFieldType;
    
  public:
    FVBaseFunction ( const int baseNum )
    : baseNum_ ( baseNum ) 
    { 
      assert( (baseNum_ >= 0) && (baseNum_ < dimRange) );
    }
    
    virtual void evaluate ( const FieldVector<int, 0> &diffVariable, 
                            const DomainType & x, RangeType & phi) const 
    {
      phi = 0;
      phi[baseNum_] = 1;
    }

    virtual void evaluate ( const FieldVector<int, 1> &diffVariable, 
                            const DomainType & x, RangeType & phi) const 
    {
      phi = 0;
    }

    virtual void evaluate ( const FieldVector<int, 2> &diffVariable, 
                            const DomainType & x, RangeType & phi) const 
    {
      phi = 0;
    }
  };



  //! Factory class for base functions
  template< class FunctionSpace, int polOrd >
  class FVBaseFunctionFactory
  : public BaseFunctionFactory< FunctionSpace >
  {
    dune_static_assert( polOrd == 0, "Only implemented for PolOrd=0." );
      
  public:
    typedef FunctionSpace FunctionSpaceType;
    typedef BaseFunctionInterface< FunctionSpaceType > BaseFunctionType;

    static const int dimRange = FunctionSpaceType::dimRange;

    FVBaseFunctionFactory ( const GeometryType &type )
    : BaseFunctionFactory< FunctionSpaceType >( type )
    {}

    virtual BaseFunctionType *baseFunction ( const int i ) const
    {
      return new FVBaseFunction< FunctionSpaceType, polOrd >( i );
    }
    
    virtual int numBaseFunctions() const 
    {
      return dimRange;
    }
  };

  } // end namespace Fem 

} // end namespace Dune

#endif // #ifndef DUNE_FVSPACEBASEFUNCTIONS_HH
