/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include "basefunctiontest.hh"

// #include <dune/grid/io/file/dgfparser/dgfgridtype.hh>
#include <dune/fem/gridpart/gridpart.hh>
#include <dune/fem/space/lagrangespace.hh>

namespace Dune {

  void LagrangeBase_Test::run() 
  {
    testBaseFunctions();
  }

  void LagrangeBase_Test::testBaseFunctions() 
  {
    typedef GridSelector::GridType GridType;
    static const int dimworld = GridSelector::dimworld;

    GridPtr< GridType > gridPtr( gridFile_ );
    GridType& grid = *gridPtr;

    typedef LeafGridPart< GridType > GridPartType; 
    GridPartType gridPart( grid );

    typedef FunctionSpace< double, double, dimworld, 1 > FunctionSpaceType;
    
    // check polynomial order 1 
    typedef LagrangeDiscreteFunctionSpace< FunctionSpaceType, GridPartType, 1 >
      OneSpaceType;
    {
      std :: cout << "Linear Base Functions" << std :: endl;
      OneSpaceType space( gridPart);
      checkLagrangeBase( space );
    }
 
    #ifdef TEST_SECOND_ORDER
    // check polynomial order 2
    typedef LagrangeDiscreteFunctionSpace< FunctionSpaceType, GridPartType, 2 >
      TwoSpaceType;
    {
      std :: cout << "Quadratic Base Functions" << std :: endl;
      TwoSpaceType space( gridPart );
      checkLagrangeBase( space );
    }
    #endif

    #ifdef TEST_THIRD_ORDER
    typedef LagrangeDiscreteFunctionSpace< FunctionSpaceType, GridPartType, 3 >
      ThreeSpaceType;
    {
      std :: cout << "Cubic Base Functions" << std :: endl;
      ThreeSpaceType space( gridPart );
      checkLagrangeBase( space );
    }
    #endif

  }
  
  template< class SpaceType > 
  void LagrangeBase_Test :: checkLagrangeBase( const SpaceType &space )
  {
    typedef typename SpaceType :: IteratorType IteratorType; 
    typedef typename SpaceType :: BaseFunctionSetType  BaseFunctionSetType;
    typedef typename SpaceType :: LagrangePointSetType LagrangePointSetType;
    typedef typename SpaceType :: GridPartType GridPartType;
    typedef typename SpaceType :: DomainType DomainType; 
    typedef typename SpaceType :: RangeType RangeType; 

    int errors = 0;
    
    IteratorType end = space.end();
    for(IteratorType it = space.begin(); it != end; ++it)
    {
      const BaseFunctionSetType& baseSet = space.baseFunctionSet( *it );
      const int numBaseFct = baseSet.numBaseFunctions();

      const LagrangePointSetType& pointSet = space.lagrangePointSet( *it );
      const int numPoints = pointSet.size();
     
      for( int i = 0; i < numPoints; ++i ) 
      {
        RangeType phi( 0.0 );
       
        const DomainType& x = pointSet.point( i );
       
        // evaluate on lagrange point 
        baseSet.evaluate( i , x , phi ); 
        if( std :: abs( phi[ 0 ] - 1.0 ) >= 1e-10 )
        {
          std :: cout << "Base function " << i << " failed at " << x
                      << " (" << phi[ 0 ] << " != 1)!" << std :: endl;
          ++errors;
        }
        
        for( int j = 0; j < numBaseFct; ++j ) 
        {
          if( i == j )
            continue;

          // evaluate on lagrange point 
          baseSet.evaluate( j , x, phi ); 
          if( std :: abs( phi[ 0 ] ) >= 1e-10 )
          {
            std :: cout << "Base function " << j << " failed at " << x
                        << " (" << phi[ 0 ] << " != 0)!" << std :: endl;
            ++errors;
          }
        }
      }
    }

    assert( errors == 0 );
  }

} // end namespace Dune
