/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#include <config.h>

#include <dune/common/forloop.hh>

#include "../basefunctions.hh"
#include "../lagrangepoints.hh"

using namespace Dune;

#define VERBOSITY_LEVEL 0

#ifndef DIMENSION
#define DIMENSION 2
#endif

#ifndef POLORDER
#define POLORDER 2
#endif

#ifndef GEOMETRYTYPE
#define GEOMETRYTYPE simplex
#endif

typedef FunctionSpace< double, double, DIMENSION, 1 > FunctionSpaceType;

typedef FunctionSpaceType :: DomainType DomainType;
typedef FunctionSpaceType :: RangeType RangeType;

typedef Fem::LagrangeBaseFunctionFactory< FunctionSpaceType, DIMENSION, POLORDER >
  BaseFunctionFactoryType;
typedef BaseFunctionInterface< FunctionSpaceType > BaseFunctionType;

typedef Fem::LagrangePoint< GeometryType :: GEOMETRYTYPE, DIMENSION, POLORDER >
  LagrangePointType;
typedef Fem::LagrangePointListImplementation
  < double, GeometryType :: GEOMETRYTYPE, DIMENSION, POLORDER >
  LagrangePointListType;



template< int codim >
struct PrintMaxDofs
{
  static void apply ()
  {
    const unsigned int maxCodimDofs = LagrangePointType::template Codim< codim >::maxDofs();
    std::cout << "MaxDofs< " << codim << " >: " << maxCodimDofs << std::endl;
  }
};



int main( int argc, char **argv )
{
  GeometryType geometryType( GeometryType::GEOMETRYTYPE, DIMENSION );
  BaseFunctionFactoryType baseFunctionFactory ( geometryType );

  const unsigned int numBaseFunctions = baseFunctionFactory.numBaseFunctions();

  std::cout << "Number of base functions: " << numBaseFunctions;
  std::cout << std::endl << std::endl;

  ForLoop< PrintMaxDofs, 0, DIMENSION >::apply();
  std::cout << std::endl;

  unsigned int errors = 0;
  unsigned int indexErrors = 0;
  unsigned int pointSetErrors = 0;
  LagrangePointListType pointSet( 0 );
  for( unsigned int i = 0; i < numBaseFunctions; ++i )
  {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );
    x -= pointSet.point( i );
    if( x.two_norm() >= 1e-14 )
      ++pointSetErrors;
    
    unsigned int codim, setCodim;
    unsigned int subEntity, setSubEntity;
    unsigned int dofNumber, setDofNumber;
    point.dofSubEntity( codim, subEntity, dofNumber );
    pointSet.dofSubEntity( i, setCodim, setSubEntity, setDofNumber );
    if( (setCodim != codim) || (setSubEntity != subEntity)
        || (setDofNumber != dofNumber) )
      ++pointSetErrors;
    
    unsigned int numDofs = point.numDofs( codim, subEntity );
    if( numDofs != pointSet.numDofs( codim, subEntity ) )
      ++pointSetErrors;
    
    unsigned int entityDofNumber
      = point.entityDofNumber( codim, subEntity, dofNumber );
    if( entityDofNumber != pointSet.entityDofNumber( codim, subEntity, dofNumber ) )
      ++pointSetErrors;

    if( dofNumber >= numDofs )
      ++errors;
    if( entityDofNumber != i )
      ++indexErrors;

    std :: cout << i << ": x = " << pointSet.point( i );
    std :: cout << ", height = " << point.height();
    std :: cout << ", codim = " << codim << ", subEntity = " << subEntity;
    std :: cout << ", dofNumber = " << dofNumber << " ( " << numDofs << " )";
    std :: cout << ", entityDofNumber = " << entityDofNumber;
    std :: cout << std :: endl;
  }
  std :: cout << std :: endl;
  std :: cout << "DoFs with dofNumber >= numDofs: " << errors << std :: endl;
  std :: cout << "DoFs entityDofNumber( dofSubEntity ) != i: " << indexErrors
              << std :: endl;
  std :: cout << "Errors in point set: " << pointSetErrors << std :: endl; 
  std :: cout << std :: endl;
  
  errors = 0;
  for( unsigned int j = 0; j < numBaseFunctions; ++j )
  {
    BaseFunctionType &baseFunction = *(baseFunctionFactory.baseFunction( j ));

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Base function: " << j << std :: endl;
    #endif
    for( unsigned int i = 0; i < numBaseFunctions; ++i )
    {
      LagrangePointType point( i );

      DomainType x;
      point.local( x );

      RangeType phi;
      FieldVector< int, 0 > derivative;
      baseFunction.evaluate( derivative, x, phi );
      
      double expected = ((i == j) ? 1.0 : 0.0);
      if( fabs( phi[ 0 ] - expected ) > 1e-8 )
        ++errors;

      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << i << ": x = " << x;
        std :: cout << ", f( x ) = " << phi[ 0 ];
        std :: cout << std :: endl;
      #endif
    }

    delete &baseFunction;
    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Base function evaluation errors: " << errors << std :: endl;
  
  errors = 0;
  for( unsigned int i = 0; i < numBaseFunctions; ++i ) {
    LagrangePointType point( i );

    DomainType x;
    point.local( x );

    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << "Lagrange point " << i << ": " << x << std :: endl;
    #endif

    for( int k = 0; k < DIMENSION; ++k ) {
      FieldVector< int, 1 > derivative( k );

      RangeType sum( 0 );
      for( unsigned int j = 0; j < numBaseFunctions; ++j )
      {
        BaseFunctionType &baseFunction = *(baseFunctionFactory.baseFunction( j ));

        RangeType phi;
        baseFunction.evaluate( derivative, x, phi );
        #if (VERBOSITY_LEVEL >= 1)
          std :: cout << "BaseFunction " << j << ", derivative " << k << ": " << phi[ 0 ] << std :: endl;
        #endif
        sum += phi;

        delete &baseFunction;
      }

      #if (VERBOSITY_LEVEL >= 1)
        std :: cout << "Derivative " << k << " sums up to " << sum[ 0 ] << "." << std :: endl;
      #endif
      if( fabs( sum[ 0 ] ) > 1e-8 )
        ++errors;
    }
    #if (VERBOSITY_LEVEL >= 1)
      std :: cout << std :: endl;
    #endif
  }
  std :: cout << "Base function derivtive summation errors: " << errors << std :: endl;
  
  return 0;
}
