/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_REDUCEDBASISSPACE_MAPPER_HH
#define DUNE_FEM_REDUCEDBASISSPACE_MAPPER_HH

#include <dune/fem/space/mapper/dofmapper.hh>

namespace Dune
{

  namespace Fem 
  {

  template< class GridPart, class BaseFunctionList >
  class ReducedBasisMapper;



  template< class GridPart, class BaseFunctionList >
  struct ReducedBasisMapperTraits
  {
    typedef ReducedBasisMapper< GridPart, BaseFunctionList > DofMapperType;

    typedef GridPart GridPartType;
    typedef BaseFunctionList BaseFunctionListType;
    
    typedef typename GridPartType::template Codim< 0 >::EntityType ElementType;
    
    typedef DefaultDofMapIterator< ElementType, DofMapperType > DofMapIteratorType;
  };



  /** \class ReducedBasisMapper
   *  \brief provides the mapper for the reduced basis space 
   *
   *  This mapper just performs the identity mapping.
   */
  template< class GridPart, class BaseFunctionList >
  class ReducedBasisMapper
  : public DofMapperDefault< ReducedBasisMapperTraits< GridPart, BaseFunctionList > >
  {
    typedef ReducedBasisMapper< GridPart, BaseFunctionList > ThisType;
    typedef DofMapperDefault< ReducedBasisMapperTraits< GridPart, BaseFunctionList > > BaseType;

  public:
    typedef typename BaseType::Traits Traits;

    typedef typename BaseType::ElementType ElementType;
    typedef typename BaseType::DofMapIteratorType DofMapIteratorType;
   
    typedef typename Traits::GridPartType GridPartType;
    typedef typename Traits::BaseFunctionListType BaseFunctionListType;

  public:
    explicit ReducedBasisMapper ( const BaseFunctionListType &baseFunctionList )
    : baseFunctionList_( baseFunctionList )
    {}

    /** \copydoc Dune::DofMapper::begin(const ElementType &entity) const */
    DofMapIteratorType begin ( const ElementType &entity ) const
    {
      return DofMapIteratorType( DofMapIteratorType::beginIterator, entity, *this );
    }
    
    /** \copydoc Dune::DofMapper::end(const ElementType &entity) const */
    DofMapIteratorType end ( const ElementType &entity ) const
    {
      return DofMapIteratorType( DofMapIteratorType::endIterator, entity, *this );
    }

    /** \copydoc Dune::DofMapper::mapEach */
    template< class Functor >
    void mapEach ( const ElementType &element, Functor f ) const
    {
      const int n = size();
      for( int i = 0; i < n; ++i )
        f( i, i );
    }

    /** \copydoc Dune::DofMapper::mapToGlobal(const ElementType &entity,const int localDof) const */
    int mapToGlobal ( const ElementType &entity, const int localDof ) const
    {
      return localDof;
    }

    /** \copydoc Dune::DofMapper::mapEntityDofToGlobal(const Entity &entity,const int localDof) const */
    template< class Entity >
    int mapEntityDofToGlobal ( const Entity &entity, const int localDof ) const
    {
      DUNE_THROW( NotImplemented, "ReducedBasisSpace cannot map entity DoFs." );
      return 0;
    }

    /** \copydoc Dune::DofMapper::numDofs(const ElementType &element) const */
    int numDofs ( const ElementType &element ) const
    {
      return size();
    }

    /** \copydoc Dune::DofMapper::maxNumDofs() const */
    int maxNumDofs () const
    {
      return size();
    }

    /** \copydoc Dune::DofMapper::numEntityDofs(const Entity &entity) const */
    template< class Entity >
    int numEntityDofs ( const Entity &entity ) const
    {
      DUNE_THROW( NotImplemented, "ReducedBasisSpace cannot map entity DoFs." );
      return 0;
    }
   
    /** \copydoc Dune::DofMapper::consecutive() const */
    bool consecutive () const
    {
      return false;
    }

    /** \copydoc Dune::DofMapper::newIndex(const int hole,const int block) const */
    int newIndex ( const int hole, const int block ) const
    {
      return -1;
    }

    /** \copydoc Dune::DofMapper::numberOfHoles(const int block) const */
    int numberOfHoles ( const int block ) const
    {
      return 0;
    }

    /** \copydoc Dune::DofMapper::oldIndex(const int hole,const int block) const */
    int oldIndex ( const int hole, const int block ) const
    {
      return -1;
    }

    /** \copydoc Dune::DofMapper::size() const */
    int size () const
    {
      return baseFunctionList_.size();
    }

  protected:
    const BaseFunctionListType &baseFunctionList_;
  };

  } // namespace Fem
  
} // namespace Dune

#endif // #ifndef DUNE_FEM_REDUCEDBASISSPACE_MAPPER_HH
