/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#ifndef DUNE_FEM_REDUCEDBASISSPACE_SINESPACE_HH
#define DUNE_FEM_REDUCEDBASISSPACE_SINESPACE_HH

#include <dune/fem/space/reducedbasisspace.hh>
#include <dune/fem/function/adaptivefunction.hh>
#include <dune/fem/operator/lagrangeinterpolation.hh>

namespace Dune
{

  template< class FunctionSpace >
  class SineBaseFunction
  : public Fem::Function< FunctionSpace, SineBaseFunction< FunctionSpace > >
  {
    typedef SineBaseFunction< FunctionSpace > ThisType;
    typedef Fem::Function< FunctionSpace, ThisType > BaseType;

  public:
    typedef FunctionSpace FunctionSpaceType;

    typedef typename FunctionSpaceType::DomainType DomainType;
    typedef typename FunctionSpaceType::RangeType RangeType;

    typedef typename FunctionSpaceType::DomainFieldType DomainFieldType;
    typedef typename FunctionSpaceType::RangeFieldType RangeFieldType;

    static const int dimDomain = FunctionSpaceType::dimDomain;
    static const int dimRange = FunctionSpaceType::dimRange;

    typedef FieldVector< int, dimDomain > CoefficientType;
    
    explicit SineBaseFunction ( const CoefficientType coefficient )
    : coefficient_( coefficient )
    {}

    void evaluate ( const DomainType &x, RangeType &y ) const
    {
      y = 1;
      for( int i = 0; i < dimDomain; ++i )
      {
        /*
        if( coefficient_[ i ] < 0 )
          y *= sqrt( 2 ) * cos( 2 * M_PI * coefficient_[ i ] * x[ i ] );
        else if( coefficient_[ i ] > 0 )
          y *= sqrt( 2 ) * sin( 2 * M_PI * coefficient_[ i ] * x[ i ] );
        */
        y *= sqrt( 2 ) * sin( M_PI * coefficient_[ i ] * x[ i ] );
      }
    }

    void evaluate ( const DomainType &x, const RangeFieldType t, RangeType &y ) const
    {
      evaluate( x, y );
    }

  protected:
    const CoefficientType coefficient_;
  };



  template< class BaseFunctionSpaceImp, unsigned int maxCoefficient >
  class SineReducedBasisSpace
  : public ReducedBasisSpace< AdaptiveDiscreteFunction< BaseFunctionSpaceImp > >
  {
    typedef SineReducedBasisSpace< BaseFunctionSpaceImp, maxCoefficient > ThisType;
    typedef ReducedBasisSpace< AdaptiveDiscreteFunction< BaseFunctionSpaceImp > > BaseType;

  public:
    typedef BaseFunctionSpaceImp BaseFunctionSpaceType;

    typedef AdaptiveDiscreteFunction< BaseFunctionSpaceType > BaseFunctionType;

    typedef SineBaseFunction< typename BaseFunctionSpaceType::FunctionSpaceType >
      ContinuousBaseFunctionType;

  private:
    typedef typename ContinuousBaseFunctionType::CoefficientType CoefficientType;

  public:
    explicit SineReducedBasisSpace ( BaseFunctionSpaceType &baseFunctionSpace )
    : BaseType( baseFunctionSpace )
    {
      // CoefficientType coefficient( -maxCoefficient );
      CoefficientType coefficient( 1 );
      while( true )
      {
        addBaseFunction( coefficient );

        ++coefficient[ 0 ];
        for( unsigned int i = 0; coefficient[ i ] > (int)maxCoefficient; ++i )
        {
          // coefficient[ i ] = -maxCoefficient;
          coefficient[ i ] = 1;
          if( i+1 < CoefficientType :: dimension )
            ++coefficient[ i+1 ];
          else
            return;
        }
      }
    }

  private:
    static int abs( const CoefficientType &coefficient )
    {
      int value = 0;
      for( unsigned int i = 0; i < CoefficientType :: dimension; ++i )
        value += (coefficient[ i ] < 0 ? -coefficient[ i ] : coefficient[ i ]);
      return value;
    }
    
    void addBaseFunction( const CoefficientType &coefficient )
    {
      BaseFunctionType discreteBaseFunction( "base function", baseFunctionSpace_ );
      ContinuousBaseFunctionType continuousBaseFunction( coefficient );
      LagrangeInterpolation< BaseFunctionType >
        ::interpolateFunction( continuousBaseFunction, discreteBaseFunction );
      BaseType::addBaseFunction( discreteBaseFunction );
    }

    using BaseType::baseFunctionSpace_;
  };

}

#endif // #ifndef DUNE_FEM_REDUCEDBASISSPACE_SINESPACE_HH
