/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#undef NDEBUG

#include <config.h>
#include <iostream>

#include <dune/fem/gridpart/hierarchicgridpart.hh>
#include <dune/fem/function/adaptivefunction.hh>

#include "testgrid.hh"
#include "dfspace.hh"


template< class EntityType, class MapperType >
inline bool checkIterator ( const EntityType &entity,
                            const MapperType &mapper )
{
  typedef typename MapperType :: DofMapIteratorType DofMapIteratorType;

  int count = 0;
  const DofMapIteratorType end = mapper.end( entity );
  for( DofMapIteratorType it = mapper.begin( entity ); it != end; ++it )
  {
    if( it.global() != mapper.mapToGlobal( entity, it.local() ) )
      return false;
    ++count;
  }
  if( count != mapper.numDofs( entity ) )
    return false;

  return true;
}


template< class DiscreteFunctionSpaceType >
inline bool checkMappers ( const DiscreteFunctionSpaceType &space )
{
  typedef typename DiscreteFunctionSpaceType :: MapperType MapperType;
  typedef typename DiscreteFunctionSpaceType :: BlockMapperType BlockMapperType;
  typedef typename DiscreteFunctionSpaceType :: IteratorType IteratorType;

  typedef typename IteratorType :: Entity EntityType;
  
  enum { blockSize = DiscreteFunctionSpaceType :: localBlockSize };

  const MapperType &mapper = space.mapper();
  const BlockMapperType &blockMapper = space.blockMapper();

  const IteratorType end = space.end();
  for( IteratorType it = space.begin(); it != end; ++it )
  {
    const EntityType &entity = *it;

    if( !checkIterator( entity, mapper ) )
    {
      std :: cout << "Mapper: DoF Iterator is inconsistent with DoF mapping."
                  << std :: endl;
      return false;
    }
    
    if( !checkIterator( entity, blockMapper ) )
    {
      std :: cout << "BlockMapper: DoF Iterator is inconsistent with DoF mapping."
                  << std :: endl;
      return false;
    }

    const int numDofs = mapper.numDofs( entity );
    const int numDofBlocks = blockMapper.numDofs( entity );
    /*
    std :: cout << "numDofs: " << numDofs
                << ", numDofBlocks: " << numDofBlocks
                << ", blockSize: " << blockSize
                << std :: endl;
    */
    if( numDofs != blockSize * numDofBlocks )
    {
      std :: cerr << "Number of DoFs is inconsistent between DoF mappers."
                  << std :: endl;
      return false;
    }

    for( int block = 0; block < numDofBlocks; ++block )
    {
      const int blockGlobal = blockMapper.mapToGlobal( entity, block );
      for( int i = 0; i < blockSize; ++i )
      {
        const int dof = blockSize * block + i;
        const int dofGlobal = mapper.mapToGlobal( entity, dof );

#if 0        
        std :: cout << "local DoF: " << dof << ", globalDoF: " << dofGlobal
                    << ", blockMapped: " << (blockSize * blockGlobal + i)
                    << std :: endl;
#endif

        if( dofGlobal != blockSize * blockGlobal + i )
        {
          std :: cerr << "DoF mapping inconsistent between DoF mappers."
                      << std :: endl;
          return false;
        }
      }
    }
  }

  return true;
}


using namespace Dune;

typedef GridSelector::GridType MyGridType;
typedef HierarchicGridPart< MyGridType > GridPartType;

typedef TestFunctionSpace FunctionSpaceType;
typedef TestDiscreteFunctionSpace< GridPartType > DiscreteFunctionSpaceType;


int main ( int argc, char *argv[] )
try
{
  MPIManager :: initialize( argc, argv );

  MyGridType &grid = TestGrid :: grid();
  // const int step = TestGrid :: refineStepsForHalf();
  // grid.globalRefine( 2*step );

  GridPartType gridPart( grid );
  DiscreteFunctionSpaceType discreteFunctionSpace( gridPart );

  AdaptiveDiscreteFunction< DiscreteFunctionSpaceType >
    discreteFunction( "test function",discreteFunctionSpace );

  int ret = (checkMappers( discreteFunctionSpace ) ? 0 : 1);

  Dune::GlobalRefine::apply(grid,Dune::DGFGridInfo< MyGridType >::refineStepsForHalf());
  ret += (checkMappers( discreteFunctionSpace ) ? 0 : 2);

  return ret;
}
catch( Exception e )
{
  std :: cerr << e.what() << std :: endl;
  return 1;
}
