/**************************************************************************
 
  The dune-fem module is a module of DUNE (see www.dune-project.org).
  It is based on the dune-grid interface library 
  extending the grid interface by a number of discretization algorithms
  for solving non-linear systems of partial differential equations.

  Copyright (C) 2003 - 2012 Robert Kloefkorn
  Copyright (C) 2003 - 2010 Mario Ohlberger 
  Copyright (C) 2004 - 2012 Andreas Dedner
  Copyright (C) 2005 - 2010 Mirko Kraenkel
  Copyright (C) 2006 - 2012 Martin Nolte
  Copyright (C) 2011 - 2012 Tobias Malkmus

  The dune-fem module is free software; you can redistribute it and/or 
  modify it under the terms of the GNU General Public License as 
  published by the Free Software Foundation; either version 2 of 
  the License, or (at your option) any later version.

  The dune-fem module is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 
**************************************************************************/
#undef NDEBUG

#include <config.h>
#include <iostream>

#include <dune/fem/gridpart/hierarchicgridpart.hh>
#include <dune/fem/misc/mpimanager.hh>
#include <dune/fem/operator/projection/vtxprojection.hh>
#include <dune/fem/operator/projection/l2projection.hh>
#include <dune/fem/operator/lagrangeinterpolation.hh>

#if defined USE_BLOCKVECTORFUNCTION
#include <dune/fem/function/blockvectorfunction.hh>
#elif defined USE_VECTORFUNCTION
#include <dune/fem/storage/vector.hh>
#include <dune/fem/function/vectorfunction.hh>
#elif defined USE_ATTACHEDFUNCTION
#include <dune/fem/function/attachedfunction/function.hh>
#else
#include <dune/fem/function/adaptivefunction.hh>
#endif

#include "testgrid.hh"
#include "dfspace.hh"
// #include "vtxl2projection.hh"
#include "exactsolution.hh"
#include <dune/fem/gridpart/adaptiveleafgridpart.hh>

using namespace Dune;

typedef GridSelector::GridType MyGridType;
// typedef HierarchicGridPart< MyGridType > GridPartType;
typedef Fem::AdaptiveLeafGridPart< MyGridType > GridPartType;

typedef TestFunctionSpace FunctionSpaceType;
typedef TestDiscreteFunctionSpace< GridPartType > DiscreteFunctionSpaceType;

#if defined USE_BLOCKVECTORFUNCTION
typedef Fem :: ISTLBlockVectorDiscreteFunction< DiscreteFunctionSpaceType >
  DiscreteFunctionType;
#elif defined USE_VECTORFUNCTION
typedef Fem :: ManagedDiscreteFunction
  < Fem::VectorDiscreteFunction
    < DiscreteFunctionSpaceType,
      DynamicVector< FunctionSpaceType :: RangeFieldType > > >
  DiscreteFunctionType;
#elif defined USE_ATTACHEDFUNCTION
typedef Fem :: AttachedDiscreteFunction< DiscreteFunctionSpaceType >
  DiscreteFunctionType;
#else
typedef Fem :: AdaptiveDiscreteFunction< DiscreteFunctionSpaceType >
  DiscreteFunctionType;
#endif

typedef ExactSolution< FunctionSpaceType > ExactSolutionType;

#include <dune/fem/io/file/vtkio.hh>

// main program 
int main(int argc, char ** argv) 
{
  MPIManager :: initialize( argc, argv );
  try
  {
    MyGridType &grid = TestGrid :: grid();
    // const int step = TestGrid :: refineStepsForHalf();
    // grid.globalRefine( 2*step );

    GridPartType gridPart( grid );
    DiscreteFunctionSpaceType discreteFunctionSpace( gridPart ); // DG
    typedef LagrangeDiscreteFunctionSpace<FunctionSpaceType, GridPartType,
      1,CachingStorage> LDiscreteFunctionSpaceType; 
    LDiscreteFunctionSpaceType lagspace(gridPart);
    ExactSolutionType exactSolution;
    typedef AdaptiveDiscreteFunction<LDiscreteFunctionSpaceType> LagrangeFunctionType;

    // perform the L2Projection
    DiscreteFunctionType solution( "solution", discreteFunctionSpace );
    Fem::L2Projection< ExactSolutionType,  DiscreteFunctionType > dgl2; 
    dgl2( exactSolution, solution );
    
    LagrangeFunctionType contSolution("contSolution",lagspace);
    VtxProjection< DiscreteFunctionType,LagrangeFunctionType > projection;
    projection(solution,contSolution);

    LagrangeFunctionType lagrangeSolution("lagrangeSolution",lagspace);
    LagrangeInterpolation< LagrangeFunctionType >
      :: interpolateFunction( exactSolution, lagrangeSolution );
    #if 0
    LagrangeType lagrangeContSolution("lagrnageContSolution",lagspace);
    VtxProjection<double,double,LagrangeType,LagrangeType> projection2;
    projection2(lagrangeSolution,lagrangeContSolution);

    // output to vtk file
    VTKIO<GridPartType> vtkWriter(gridPart);
    vtkWriter.addCellData(solution);
    vtkWriter.addVertexData(contSolution);
    lagrangeSolution -= lagrangeContSolution;
    vtkWriter.addVertexData(lagrangeSolution);
    vtkWriter.addVertexData(lagrangeContSolution);
    vtkWriter.pwrite("vtxprojection",
		     Parameter::commonOutputPath().c_str(),".",
		     Dune::VTKOptions::ascii);

#endif
    return 0;
  }
  catch( Exception e )
  {
    std :: cerr << e.what() << std :: endl;
    return 1;
  }
}
